package dev.fitko.fitconnect.sdk.tools;

import com.nimbusds.jose.JWEAlgorithm;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.KeyType;
import org.hamcrest.CoreMatchers;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.Map;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.notNullValue;
import static org.hamcrest.Matchers.nullValue;

class JWKGeneratorTest {

    private JWKGenerator underTest;

    @BeforeEach
    void setup() {
        underTest = new JWKGenerator();
    }

    @Test
    void testPublicEncryptionKey() {

        // When
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(2048);

        final JWK publicKey = encryptionKeyPair.getPublicKey();

        // Then
        assertThat(publicKey, CoreMatchers.is(notNullValue()));


        assertThat(publicKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(publicKey.getKeyOperations(), hasSize(1));
        assertThat(publicKey.getKeyOperations(), contains(KeyOperation.WRAP_KEY));

        assertThat(publicKey.getX509CertChain(), hasSize(1));

        assertThat(publicKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat( publicKey.getAlgorithm(), CoreMatchers.is(JWEAlgorithm.RSA_OAEP_256));
    }

    @Test
    void testPrivateDecryptionKey() {

        // When
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(2048);

        final JWK privateKey = encryptionKeyPair.getPrivateKey();

        // Then
        assertThat(privateKey, CoreMatchers.is(CoreMatchers.notNullValue()));

        assertThat(privateKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(privateKey.getKeyOperations(), hasSize(1));
        assertThat(privateKey.getKeyOperations(), contains(KeyOperation.UNWRAP_KEY));

        assertThat(privateKey.getX509CertChain(), CoreMatchers.is(nullValue()));

        assertThat(privateKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat(privateKey.getAlgorithm(), CoreMatchers.is(JWEAlgorithm.RSA_OAEP_256));

        final Map<String, Object> keyParams = privateKey.toRSAKey().toJSONObject();

        assertThat(keyParams.get("d"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dp"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dq"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("e"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("n"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("p"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("q"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("qi"), CoreMatchers.is(CoreMatchers.notNullValue()));
    }

    @Test
    void testPublicSignatureVerificationKey() {

        // When
        final JWKPair signatureKeyPair = underTest.generateSignatureKeyPair(2048);

        final JWK publicKey = signatureKeyPair.getPublicKey();

        // Then
        assertThat(publicKey, CoreMatchers.is(CoreMatchers.notNullValue()));

        assertThat(publicKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(publicKey.getKeyOperations(), hasSize(1));
        assertThat(publicKey.getKeyOperations(), contains(KeyOperation.VERIFY));

        assertThat(publicKey.getX509CertChain(), hasSize(1));

        assertThat(publicKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat( publicKey.getAlgorithm(), CoreMatchers.is(JWSAlgorithm.PS512));

    }

    @Test
    void testPrivateSigningKey() {

        // When
        final JWKPair signatureKeyPair = underTest.generateSignatureKeyPair(2048);

        final JWK privateKey =  signatureKeyPair.getPrivateKey();

        // Then
        assertThat(privateKey, CoreMatchers.is(notNullValue()));


        assertThat(privateKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(privateKey.getKeyOperations(), hasSize(1));
        assertThat(privateKey.getKeyOperations(), contains(KeyOperation.SIGN));

        assertThat(privateKey.getX509CertChain(), CoreMatchers.is(nullValue()));

        assertThat(privateKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat(privateKey.getAlgorithm(), CoreMatchers.is(JWSAlgorithm.PS512));

        final Map<String, Object> keyParams = privateKey.toRSAKey().toJSONObject();

        assertThat(keyParams.get("d"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dp"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dq"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("e"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("n"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("p"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("q"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("qi"), CoreMatchers.is(CoreMatchers.notNullValue()));

    }
}